#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "keymap.h"
#include "translate.h"
#include <linux/vt.h>

#define ESC 27

int shifts[16]={0,0,0,0,0,0,0,0,0,0,0,0};
int keys_down[128];
int tty,psaux;
int ttynum;
char ttyname[24];
static int npadch = -1;			/* -1 or number assembled on pad */
int console_control;

static char at2xt[ 256 ] = {
/*         0    1    2    3    4    5    6    7    8    9    a    b    c    d    e    f */
/* 0 */ 0x00,0x43,0x00,0x3f,0x3d,0x3b,0x3c,0x58,0x00,0x44,0x42,0x40,0x3e,0x0f,0x29,0x00,
/* 1 */ 0x00,0x38,0x2a,0x00,0x1d,0x10,0x02,0x00,0x00,0x00,0x2c,0x1f,0x1e,0x11,0x03,0x5b,
/* 2 */ 0x00,0x2e,0x2d,0x20,0x12,0x05,0x04,0x5c,0x00,0x39,0x2f,0x21,0x14,0x13,0x06,0x5d,
/* 3 */ 0x00,0x31,0x30,0x23,0x22,0x15,0x07,0x6a,0x00,0x00,0x32,0x24,0x16,0x08,0x09,0x69,
/* 4 */ 0x00,0x33,0x25,0x17,0x18,0x0b,0x0a,0x00,0x00,0x34,0x35,0x26,0x27,0x19,0x0c,0x00,
/* 5 */ 0x00,0x00,0x28,0x00,0x1a,0x0d,0x00,0x00,0x3a,0x36,0x1c,0x1b,0x00,0x2b,0x68,0x00,
/* 6 */ 0x00,0x00,0x00,0x00,0x00,0x00,0x0e,0x00,0x00,0x4f,0x00,0x4b,0x47,0x00,0x00,0x00,
/* 7 */ 0x52,0x53,0x50,0x4c,0x4d,0x48,0x01,0x45,0x57,0x4e,0x51,0x4a,0x37,0x49,0x46,0x00,
/* 8 */ 0x00,0x00,0x00,0x41,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* 9 */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* a */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* b */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* c */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* d */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* e */ 0xe0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
/* f */ 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 
};

int is_a_console(int fd) 
{
  char arg;
  
  arg = 0;
  return (ioctl(fd, KDGKBTYPE, &arg) == 0
	  && ((arg == KB_101) || (arg == KB_84)));
}

static int open_a_console(char *fnam) 
{
  int fd;
  
  fd = open(fnam, O_RDONLY);
  
  /* if failed, fail */
  if (fd < 0)
      return -1;
  
  /* if not a console, fail */
  if (! is_a_console(fd))
    {
      close(fd);
      return -1;
    }
  
  /* success */
  return fd;
}

int get_console_fd(void) 
{
  int fd;

  fd = open_a_console("/dev/tty");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/tty0");
  if (fd >= 0)
    return fd;
  
  fd = open_a_console("/dev/console");
  if (fd >= 0)
    return fd;
  
  for (fd = 0; fd < 3; fd++)
    if (is_a_console(fd))
      return fd;
  
  fprintf(stderr, 
          "Couldnt get a file descriptor referring to the console\n");
  return -1;		/* total failure */
}

void send(int value)
{
    if(ioctl(tty,TIOCSTI,&value)) {
       close(tty);
       tty=open(ttyname,O_RDONLY);
       ioctl(tty,TIOCSTI,&value);
    }
};

int raw_handle_key(int key)
{
   send(key);
}

int do_console(c) {
   int fd;
   int newtty;
   struct vt_stat vts;
   
   if((ttynum==c)||!console_control)return 0;
   ttynum=c;
   snprintf(ttyname,22,"/dev/tty%d",ttynum);
    
   newtty=open(ttyname,O_RDONLY);

   if((newtty!=-1)){
      close(tty);
      tty=newtty;
      if(console_control==3)return 0;
      fd=get_console_fd();
      ioctl(fd,VT_GETSTATE,&vts);
      ioctl(fd,VT_ACTIVATE,c);
      ioctl(fd,VT_WAITACTIVE,c);
      if((c>12)&&(c<25)&&(vts.v_active<13)&&(console_control==2)) {
         ioctl(fd,VT_ACTIVATE,vts.v_active);
         ioctl(fd,VT_WAITACTIVE,vts.v_active);
      }
      if(fd>3)close(fd);
   }
}

do_cursor(int c) {
   static const char *cur_chars = "BDCA";
   send(ESC);
   send('[');
   send(cur_chars[c]); 
}

do_fn(int c) {
   char *p;

   p=func_table[c];
   if(p) do {send(*p); p++; } while(*p);
}

int cooked_handle_key(int key) {
   int i;
   int value,v;
   int leds, alt_ch;
   int rep;
   char *p, output[2];
   static const char *pad_chars = "0123456789+-*/\015,.?()";
   static ushort *map;
   int kmp,j;

/* first handle shifts */
   leds=0;
   alt_ch=0;
   rep=0;
   if(key<128) {
      if(keys_down[key])rep=1;
      keys_down[key]=1; 
   } else keys_down[key-128]=0;
         
   kmp=0;
   for(j=0;j<4;j++)kmp+=shifts[j]?1<<j:0;
   kmp ^=shifts[12]?2:0;
   map=key_maps[kmp];
    
   if(!map)return -1;
    
   value=map[key];
   v=value&0xff;
   if(key<0x80) { 
      switch((value>>8)&0x0f){
           case 0x0b:
           case 0x00:
              if (((v&0xdf)>='A')&&((v&0xdf)<='Z')&&shifts[14])
                  v^=32;
              if(v)send(v);
              break;
           case 0x01:
              do_fn(v);
              break;
           case 0x02:
              switch(v) {
                 case 0x01: send(0x0d); break;
                 case 0x07: if(!rep) {
                               shifts[14]=1-shifts[14]; 
                               leds=1; 
                            }
                      break;
                 case 0x08: if(!rep) {
                               shifts[13]=1-shifts[13]; 
                               leds=1; 
                            }
                      break;
              }
              break;
           case 0x03: /* numeric pad */
              if(shifts[13]){
                 send(pad_chars[v]);
              } else {
                 switch(v){
                    case 2: do_cursor(0); break; 
                    case 4: do_cursor(1); break; 
                    case 6: do_cursor(2); break; 
                    case 8: do_cursor(3); break; 
                    case 0: do_fn(21); break;
                    case 1: do_fn(23); break;
                    case 3: do_fn(25); break;
                    case 5: send(ESC); send('['); send('G'); break;
                    case 7: do_fn(20); break;
                    case 9: do_fn(24); break;
                    case 16: do_fn(22); break;
                    default: send(pad_chars[v]);
                 }
              }
              break;
           case 0x04: 
                    
              break;
           case 0x05: 
              do_console(v+1);      
              break;
           case 0x06:
              do_cursor(v);
              break;
           case 0x07: if(!rep)
              shifts[v]+=1;       
              break;
           case 0x08: 
                    
              break;
           case 0x09: { /* ascii */ 
 	      int base;
                
	      if (v < 10)    /* decimal input of code, while Alt depressed */
                 base = 10;
              else {       /* hexadecimal input of code, while AltGr depressed */
	         v -= 10; base = 16;
              }
	      if (npadch == -1) npadch = value; else npadch = npadch * base + value;
              }
              break;
           case 0x0a: 
              switch(v) {
                 case 0x01: if(!rep) {
                               shifts[12]=1-shifts[12]; 
                               leds=1; 
                            }
                      break;
              }
              break;
           case 0x0c: 
                   
              break;
           case 0x0d: 
                 
              break;
      }
   } else {
      switch((value>>8)&0x0f){
           case 0x07:
              if(shifts[v])shifts[v]-=1;       
              break;
      }
   }
   
   if(leds){ 
      output[0]=0xed;
      output[1]=shifts[12]+(shifts[13]<<1)+(shifts[14]<<2);
      write(psaux,output,2);
   }
   
   if((alt_ch)&&(npadch!=-1)) {
      send(npadch);
      npadch=-1;
   }
   
}

int main (int argc, char *argv[])
{
    unsigned char ch,c2,c3,c4;
    int brk=0;
    int raw,leds;
    int kbmode;
    int i;
    char output[2];
    
    if((argc<2)||(argc>3)||((argc==3)&&
       ((strcmp(argv[1],"-n"))&&(strcmp(argv[1],"-c"))
         &&(strcmp(argv[1],"-k")))) ){
       printf("Usage: ps2key [options] <tty_num>\n\n"
       "options:\n"
       "\t-n\tnormal console switching\n"
       "\t-c\tno console switching\n"
       "\t-k\tno switching, but changing the console that recieves\n"
       "\t\tkeypresses is possible\n\n"
       );
       exit(1);
    }

    console_control=2;
    if(argc==3) {
       if(!strcmp(argv[1],"-n"))console_control=1;
       if(!strcmp(argv[1],"-c"))console_control=0;
       if(!strcmp(argv[1],"-k"))console_control=3;
    }
    
    ttynum=strtol(argv[argc-1],NULL,0);
    snprintf(ttyname,22,"/dev/tty%d",ttynum);
    
    psaux=open("/dev/psaux",O_RDWR);
    tty=open(ttyname,O_RDONLY);

    output[0]=0xed;
    output[1]=shifts[12]+(shifts[13]<<1)+(shifts[14]<<2);
    write(psaux,output,2);

    output[0]=0xf3;
    output[1]=0x00;
    write(psaux,output,2);
    
    for(i=0;i<128;i++)keys_down[i]=0;

    do {
       read(psaux,&ch,1);
       ioctl(tty,KDGKBMODE,&kbmode);
       if((kbmode&1)==0)raw=1; else raw=0;
       if(ch==0xe1) { /* pause */
         read(psaux,&c2,1);
         if(c2==0xf0){ 
            read(psaux,&c2,1);
            brk=0x80;
         }
         read(psaux,&c3,1);
         if(c3==0xf0){ 
            read(psaux,&c3,1);
            brk=0x80;
         }
       };
       if(ch==0xe0){
         read(psaux,&c2,1);
         if(c2==0xf0){ 
            read(psaux,&c2,1);
            brk=0x80;
         }
         if(raw) {
            send(0xe0);
            send(brk|at2xt[c2]);
            if((!brk)&&(c2==0x2f)) {
               shifts[12]=1-shifts[12];
               output[0]=0xed;
               output[1]=shifts[12]+(shifts[13]<<1)+(shifts[14]<<2);
               write(psaux,output,2);
            }
         } else cooked_handle_key(brk|e0_keys[at2xt[c2]]);
         brk=0;
       } else if (ch==0xf0) { brk=0x80;}
       else {
          if(raw) {
             send(brk|at2xt[ch]); 
             leds=0; /* follow leds status for display only. */
             if(!brk)switch(ch){
                case 0x58:shifts[14]=1-shifts[14]; leds=1; break;
                case 0x77:shifts[13]=1-shifts[13]; leds=1; break;
             };
             if(leds){ 
                output[0]=0xed;
                output[1]=shifts[12]+(shifts[13]<<1)+(shifts[14]<<2);
                write(psaux,output,2);
             };
          } else cooked_handle_key(brk|at2xt[ch]);
          brk=0;
       }
    }  while(1);
}
